"""
Interpolate Data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
from meteostat.core.warn import warn


def interpolate(self, limit: int = 3):
    """
    Interpolate NULL values
    """

    if self.count() > 0 and not self._data.isnull().values.all():

        # Create temporal instance
        temp = copy(self)

        # Apply interpolation
        temp._data = temp._data.groupby("station").apply(
            lambda group: group.interpolate(
                method="linear", limit=limit, limit_direction="both", axis=0
            )
        )

        # Return class instance
        return temp

    # Show warning & return self
    warn("Skipping interpolation on empty DataFrame")
    return self



"""
Get Number Of Rows

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""


def count(self) -> int:
    """
    Return number of rows in DataFrame
    """

    return len(self._data.index)



"""
Utilities - Endpoint Helpers

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from typing import Union
from meteostat.enumerations.granularity import Granularity


def generate_endpoint_path(
    granularity: Granularity,
    station: str,
    year: Union[int, None] = None,
    map_file: bool = False,  # Is a source map file?
) -> str:
    """
    Generate Meteostat Bulk path
    """

    # Base path
    path = f"{granularity.value}/"

    if granularity == Granularity.HOURLY and year:
        path += f"{year}/"

    appendix = ".map" if map_file else ""

    return f"{path}{station}{appendix}.csv.gz"



"""
Core Class - Data Loader

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from urllib.error import HTTPError
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
from typing import Callable, Union
import pandas as pd
from meteostat.core.warn import warn


def processing_handler(
    datasets: list, load: Callable[[dict], None], cores: int, threads: int
) -> None:
    """
    Load multiple datasets (simultaneously)
    """

    # Data output
    output = []

    # Multi-core processing
    if cores > 1 and len(datasets) > 1:

        # Create process pool
        with Pool(cores) as pool:

            # Process datasets in pool
            output = pool.starmap(load, datasets)

            # Wait for Pool to finish
            pool.close()
            pool.join()

    # Multi-thread processing
    elif threads > 1 and len(datasets) > 1:

        # Create process pool
        with ThreadPool(threads) as pool:

            # Process datasets in pool
            output = pool.starmap(load, datasets)

            # Wait for Pool to finish
            pool.close()
            pool.join()

    # Single-thread processing
    else:

        for dataset in datasets:
            output.append(load(*dataset))

    # Remove empty DataFrames
    filtered = list(filter(lambda df: df.index.size > 0, output))

    return pd.concat(filtered) if len(filtered) > 0 else output[0]


def load_handler(
    endpoint: str,
    path: str,
    columns: list,
    types: Union[dict, None],
    parse_dates: list,
    coerce_dates: bool = False,
) -> pd.DataFrame:
    """
    Load a single CSV file into a DataFrame
    """

    try:

        # Read CSV file from Meteostat endpoint
        df = pd.read_csv(
            endpoint + path,
            compression="gzip",
            names=columns,
            dtype=types,
            parse_dates=parse_dates,
        )

        # Force datetime conversion
        if coerce_dates:
            df.iloc[:, parse_dates] = df.iloc[:, parse_dates].apply(
                pd.to_datetime, errors="coerce"
            )

    except (FileNotFoundError, HTTPError):

        # Create empty DataFrane
        df = pd.DataFrame(columns=[*types])

        # Display warning
        warn(f"Cannot load {path} from {endpoint}")

    # Return DataFrame
    return df



"""
Base Interface Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import os


class Base:

    """
    Base class that provides features which are used across the package
    """

    # Base URL of the Meteostat bulk data interface
    endpoint: str = "https://bulk.meteostat.net/v2/"

    # Location of the cache directory
    cache_dir: str = os.path.expanduser("~") + os.sep + ".meteostat" + os.sep + "cache"

    # Auto clean cache directories?
    autoclean: bool = True

    # Maximum age of a cached file in seconds
    max_age: int = 24 * 60 * 60

    # Number of processes used for processing files
    processes: int = 1

    # Number of threads used for processing files
    threads: int = 1



"""
Utilities - DataFrame Mutations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
import numpy as np
import pandas as pd


def localize(df: pd.DataFrame, timezone: str) -> pd.DataFrame:
    """
    Convert time data to any time zone
    """

    return df.tz_localize("UTC", level="time").tz_convert(timezone, level="time")


def filter_time(df: pd.DataFrame, start: datetime, end: datetime) -> pd.DataFrame:
    """
    Filter time series data based on start and end date
    """

    # Get time index
    time = df.index.get_level_values("time")

    # Filter & return
    return df.loc[(time >= start) & (time <= end)]


def adjust_temp(df: pd.DataFrame, alt: int):
    """
    Adjust temperature-like data based on altitude
    """

    # Default temperature difference by 100 meters
    temp_diff = 0.6

    # Temperature-like columns
    temp_like = ("temp", "dwpt", "tavg", "tmin", "tmax")

    # Adjust values for all temperature-like data
    for col_name in temp_like:
        if col_name in df.columns:
            df.loc[df[col_name] != np.NaN, col_name] = df[col_name] + (
                temp_diff * ((df["elevation"] - alt) / 100)
            )

    return df



"""
Granularity Enumeration

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from enum import Enum


class Granularity(Enum):
    """
    The different levels of time series granularity
    """

    HOURLY = "hourly"
    DAILY = "daily"
    MONTHLY = "monthly"
    NORMALS = "normals"



"""
Hourly Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from math import floor
from datetime import datetime, timedelta
from typing import Union
import pytz
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.utilities.aggregations import degree_mean
from meteostat.interface.timeseries import TimeSeries
from meteostat.interface.point import Point


class Hourly(TimeSeries):

    """
    Retrieve hourly weather observations for one or multiple weather stations or
    a single geographical point
    """

    # The cache subdirectory
    cache_subdir: str = "hourly"

    # Granularity
    granularity = Granularity.HOURLY

    # Download data as annual chunks
    chunked: bool = True

    # The time zone
    _timezone: str = None

    # Default frequency
    _freq: str = "1H"

    # Flag which represents model data
    _model_flag = "E"

    # Raw data columns
    _columns: list = [
        "date",
        "hour",
        "temp",
        "dwpt",
        "rhum",
        "prcp",
        "snow",
        "wdir",
        "wspd",
        "wpgt",
        "pres",
        "tsun",
        "coco",
    ]

    # Index of first meteorological column
    _first_met_col = 2

    # Data types
    _types: dict = {
        "temp": "float64",
        "dwpt": "float64",
        "rhum": "float64",
        "prcp": "float64",
        "snow": "float64",
        "wdir": "float64",
        "wspd": "float64",
        "wpgt": "float64",
        "pres": "float64",
        "tsun": "float64",
        "coco": "float64",
    }

    # Columns for date parsing
    _parse_dates: dict = {"time": [0, 1]}

    # Default aggregation functions
    aggregations: dict = {
        "temp": "mean",
        "dwpt": "mean",
        "rhum": "mean",
        "prcp": "sum",
        "snow": "max",
        "wdir": degree_mean,
        "wspd": "mean",
        "wpgt": "max",
        "pres": "mean",
        "tsun": "sum",
        "coco": "max",
    }

    def _set_time(
        self, start: datetime = None, end: datetime = None, timezone: str = None
    ) -> None:
        """
        Set & adapt the period's time zone
        """

        if timezone:

            # Save timezone
            self._timezone = timezone

            if start and end:

                # Initialize time zone
                timezone = pytz.timezone(self._timezone)

                # Set start date
                start = timezone.localize(start, is_dst=None).astimezone(pytz.utc)

                # Set end date
                end = timezone.localize(end, is_dst=None).astimezone(pytz.utc)

        if self.chunked:

            self._annual_steps = [
                (start + timedelta(days=365 * i)).year
                for i in range(end.year - start.year + 2)
            ]

        self._start = start
        self._end = end

    def __init__(
        self,
        loc: Union[pd.DataFrame, Point, list, str],  # Station(s) or geo point
        start: datetime = None,
        end: datetime = None,
        timezone: str = None,
        model: bool = True,  # Include model data?
        flags: bool = False,  # Load source flags?
    ) -> None:

        # Set time zone and adapt period
        self._set_time(start, end, timezone)

        # Initialize time series
        self._init_time_series(loc, start, end, model, flags)

    def expected_rows(self) -> int:
        """
        Return the number of rows expected for the defined date range
        """

        return floor((self._end - self._start).total_seconds() / 3600) + 1



"""
Calculate Data Coverage

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""


def coverage(self, parameter: str = None) -> float:
    """
    Calculate data coverage (overall or by parameter)
    """

    if parameter is None:
        return len(self._data.index) / self.expected_rows()

    return round(self._data[parameter].count() / self.expected_rows(), 2)



"""
Meteorological Data Class

A parent class for both time series and
climate normals data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from typing import Union
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.core.cache import get_local_file_path, file_in_cache
from meteostat.core.loader import processing_handler, load_handler
from meteostat.utilities.mutations import localize, filter_time, adjust_temp
from meteostat.utilities.validations import validate_series
from meteostat.utilities.aggregations import weighted_average
from meteostat.utilities.endpoint import generate_endpoint_path
from meteostat.interface.base import Base


class MeteoData(Base):

    """
    A parent class for both time series and
    climate normals data
    """

    # The list of weather Stations
    _stations: Union[pd.Index, None] = None

    # The data frame
    _data: pd.DataFrame = pd.DataFrame()

    def _load_data(self, station: str, year: Union[int, None] = None) -> None:
        """
        Load file for a single station from Meteostat
        """

        # File name
        file = generate_endpoint_path(self.granularity, station, year)

        # Get local file path
        path = get_local_file_path(self.cache_dir, self.cache_subdir, file)

        # Check if file in cache
        if self.max_age > 0 and file_in_cache(path, self.max_age):

            # Read cached data
            df = pd.read_pickle(path)

        else:

            # Get data from Meteostat
            df = load_handler(
                self.endpoint, file, self._columns, self._types, self._parse_dates
            )

            # Validate and prepare data for further processing
            if self.granularity == Granularity.NORMALS and df.index.size > 0:
                # Add weather station ID
                # pylint: disable=unsupported-assignment-operation
                df["station"] = station

                # Set index
                df = df.set_index(["station", "start", "end", "month"])

            else:
                df = validate_series(df, station)

            # Save as Pickle
            if self.max_age > 0:
                df.to_pickle(path)

        # Localize time column
        if (
            self.granularity == Granularity.HOURLY
            and self._timezone is not None
            and len(df.index) > 0
        ):
            df = localize(df, self._timezone)

        # Filter time period and append to DataFrame
        # pylint: disable=no-else-return
        if self.granularity == Granularity.NORMALS and df.index.size > 0 and self._end:
            # Get time index
            end = df.index.get_level_values("end")
            # Filter & return
            return df.loc[end == self._end]
        elif not self.granularity == Granularity.NORMALS:
            df = filter_time(df, self._start, self._end)

        # Return
        return df

    def _get_datasets(self) -> list:
        """
        Get list of datasets
        """

        if self.granularity == Granularity.HOURLY and self.chunked:
            datasets = [
                (str(station), year)
                for station in self._stations
                for year in self._annual_steps
            ]
        else:
            datasets = [(str(station),) for station in self._stations]

        return datasets

    def _get_data(self) -> None:
        """
        Get all required data dumps
        """

        if len(self._stations) > 0:

            # Get list of datasets
            datasets = self._get_datasets()

            # Data Processings
            return processing_handler(
                datasets, self._load_data, self.processes, self.threads
            )

        # Empty DataFrame
        return pd.DataFrame(columns=[*self._types])

    # pylint: disable=too-many-branches
    def _resolve_point(
        self, method: str, stations: pd.DataFrame, alt: int, adapt_temp: bool
    ) -> None:
        """
        Project weather station data onto a single point
        """

        if self._stations.size == 0 or self._data.size == 0:
            return None

        if method == "nearest":

            if adapt_temp:

                # Join elevation of involved weather stations
                data = self._data.join(stations["elevation"], on="station")

                # Adapt temperature-like data based on altitude
                data = adjust_temp(data, alt)

                # Drop elevation & round
                data = data.drop("elevation", axis=1).round(1)

            else:

                data = self._data

            if self.granularity == Granularity.NORMALS:
                self._data = data.groupby(level=["start", "end", "month"]).agg("first")

            else:
                self._data = data.groupby(
                    pd.Grouper(level="time", freq=self._freq)
                ).agg("first")

        else:

            # Join score and elevation of involved weather stations
            data = self._data.join(stations[["score", "elevation"]], on="station")

            # Adapt temperature-like data based on altitude
            if adapt_temp:
                data = adjust_temp(data, alt)

            # Exclude non-mean data & perform aggregation
            if not self.granularity == Granularity.NORMALS:
                excluded = data["wdir"]
                excluded = excluded.groupby(
                    pd.Grouper(level="time", freq=self._freq)
                ).agg("first")

            # Aggregate mean data
            if self.granularity == Granularity.NORMALS:
                data = data.groupby(level=["start", "end", "month"]).apply(
                    weighted_average
                )

                # Remove obsolete index column
                try:
                    data = data.reset_index(level=3, drop=True)
                except IndexError:
                    pass

            else:
                data = data.groupby(pd.Grouper(level="time", freq=self._freq)).apply(
                    weighted_average
                )

                # Drop RangeIndex
                data.index = data.index.droplevel(1)

                # Merge excluded fields
                data["wdir"] = excluded

            # Drop score and elevation
            self._data = data.drop(["score", "elevation"], axis=1).round(1)

        # Set placeholder station ID
        self._data["station"] = "XXXXX"

        # Set index
        if self.granularity == Granularity.NORMALS:
            self._data = self._data.set_index("station", append=True)
            self._data = self._data.reorder_levels(["station", "start", "end", "month"])
        else:
            self._data = self._data.set_index(
                ["station", self._data.index.get_level_values("time")]
            )

        # Set station index
        self._stations = pd.Index(["XXXXX"])



"""
TimeSeries Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
from typing import Union
import numpy as np
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.core.cache import get_local_file_path, file_in_cache
from meteostat.core.loader import processing_handler, load_handler
from meteostat.utilities.mutations import localize, filter_time
from meteostat.utilities.validations import validate_series
from meteostat.utilities.endpoint import generate_endpoint_path
from meteostat.interface.point import Point
from meteostat.interface.meteodata import MeteoData


class TimeSeries(MeteoData):

    """
    TimeSeries class which provides features which are
    used across all time series classes
    """

    # The list of origin weather Stations
    _origin_stations: Union[pd.Index, None] = None

    # The start date
    _start: Union[datetime, None] = None

    # The end date
    _end: Union[datetime, None] = None

    # Include model data?
    _model: bool = True

    # Fetch source flags?
    _flags = bool = False

    def _load_flags(self, station: str, year: Union[int, None] = None) -> None:
        """
        Load flag file for a single station from Meteostat
        """

        # File name
        file = generate_endpoint_path(self.granularity, station, year, True)

        # Get local file path
        path = get_local_file_path(self.cache_dir, self.cache_subdir, file)

        # Check if file in cache
        if self.max_age > 0 and file_in_cache(path, self.max_age):

            # Read cached data
            df = pd.read_pickle(path)

        else:

            # Get data from Meteostat
            df = load_handler(
                self.endpoint,
                file,
                self._columns,
                {key: "string" for key in self._columns[self._first_met_col :]},
                self._parse_dates,
            )

            # Validate Series
            df = validate_series(df, station)

            # Save as Pickle
            if self.max_age > 0:
                df.to_pickle(path)

        # Localize time column
        if (
            self.granularity == Granularity.HOURLY
            and self._timezone is not None
            and len(df.index) > 0
        ):
            df = localize(df, self._timezone)

        # Filter time period and append to DataFrame
        if self._start and self._end:
            df = filter_time(df, self._start, self._end)

        return df

    def _get_flags(self) -> None:
        """
        Get all source flags
        """

        if len(self._stations) > 0:

            # Get list of datasets
            datasets = self._get_datasets()

            # Data Processings
            return processing_handler(
                datasets, self._load_flags, self.processes, self.threads
            )

        # Empty DataFrame
        return pd.DataFrame(columns=[*self._types])

    def _filter_model(self) -> None:
        """
        Remove model data from time series
        """

        columns = self._columns[self._first_met_col :]

        for col_name in columns:
            self._data.loc[
                (pd.isna(self._data[f"{col_name}_flag"]))
                | (self._data[f"{col_name}_flag"].str.contains(self._model_flag)),
                col_name,
            ] = np.NaN

        # Conditionally, remove flags from DataFrame
        if not self._flags:
            self._data.drop(
                map(lambda col_name: f"{col_name}_flag", columns), axis=1, inplace=True
            )

        # Drop NaN-only rows
        self._data.dropna(how="all", subset=columns, inplace=True)

    def _init_time_series(
        self,
        loc: Union[pd.DataFrame, Point, list, str],  # Station(s) or geo point
        start: datetime = None,
        end: datetime = None,
        model: bool = True,  # Include model data?
        flags: bool = False,  # Load source flags?
    ) -> None:
        """
        Common initialization for all time series, regardless
        of its granularity
        """

        # Set list of weather stations based on user
        # input or retrieve list of stations programatically
        # if location is a geographical point
        if isinstance(loc, pd.DataFrame):
            self._stations = loc.index
        elif isinstance(loc, Point):
            stations = loc.get_stations("daily", start, end, model)
            self._stations = stations.index
        else:
            if not isinstance(loc, list):
                loc = [loc]
            self._stations = pd.Index(loc)

        # Preserve settings
        self._start = start if self._start is None else self._start
        self._end = end if self._end is None else self._end
        self._model = model
        self._flags = flags

        # Get data for all weather stations
        self._data = self._get_data()

        # Load source flags through map file
        # if flags are explicitly requested or
        # model data is excluded
        if flags or not model:
            flags = self._get_flags()
            self._data = self._data.merge(
                flags, on=["station", "time"], how="left", suffixes=[None, "_flag"]
            )

        # Remove model data from DataFrame and
        # drop flags if not specified otherwise
        if not model:
            self._filter_model()

        # Interpolate data spatially if requested
        # location is a geographical point
        if isinstance(loc, Point):
            self._resolve_point(loc.method, stations, loc.alt, loc.adapt_temp)

        # Clear cache if auto cleaning is enabled
        if self.max_age > 0 and self.autoclean:
            self.clear_cache()

    # Import methods
    from meteostat.series.normalize import normalize
    from meteostat.series.interpolate import interpolate
    from meteostat.series.aggregate import aggregate
    from meteostat.series.convert import convert
    from meteostat.series.coverage import coverage
    from meteostat.series.count import count
    from meteostat.series.fetch import fetch
    from meteostat.series.stations import stations
    from meteostat.core.cache import clear_cache



"""
Core Class - Cache

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import os
import time
import hashlib


def get_local_file_path(cache_dir: str, cache_subdir: str, path: str) -> str:
    """
    Get the local file path
    """

    # Get file ID
    file = hashlib.md5(path.encode("utf-8")).hexdigest()

    return f"{cache_dir}/{cache_subdir}/{file}"


def file_in_cache(path: str, max_age: int = 0) -> bool:
    """
    Check if a file exists in the local cache
    """

    # Get directory
    directory = os.path.dirname(path)

    # Make sure the cache directory exists
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except FileExistsError:
            pass

    # Return the file path if it exists
    if os.path.isfile(path) and time.time() - os.path.getmtime(path) <= max_age:
        return True

    return False


@classmethod
def clear_cache(cls, max_age: int = None) -> None:
    """
    Clear the cache
    """

    if os.path.exists(cls.cache_dir + os.sep + cls.cache_subdir):

        # Set max_age
        if max_age is None:
            max_age = cls.max_age

        # Get current time
        now = time.time()

        # Go through all files
        for file in os.listdir(cls.cache_dir + os.sep + cls.cache_subdir):

            # Get full path
            path = os.path.join(cls.cache_dir + os.sep + cls.cache_subdir, file)

            # Check if file is older than max_age
            if now - os.path.getmtime(path) > max_age and os.path.isfile(path):
                # Delete file
                os.remove(path)



"""
Point Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
import pandas as pd
from meteostat.interface.stations import Stations


class Point:

    """
    Automatically select weather stations by geographic location
    """

    # The interpolation method (weighted or nearest)
    method: str = "nearest"

    # Maximum radius for nearby stations
    radius: int = 35000

    # Maximum difference in altitude
    alt_range: int = 350

    # Maximum number of stations
    max_count: int = 4

    # Adapt temperature data based on altitude
    adapt_temp: bool = True

    # Distance Weight
    weight_dist: float = 0.6

    # Altitude Weight
    weight_alt: float = 0.4

    # The list of weather stations
    _stations: pd.Index = None

    # The latitude
    _lat: float = None

    # The longitude
    _lon: float = None

    # The altitude
    _alt: int = None

    def __init__(self, lat: float, lon: float, alt: int = None) -> None:

        self._lat = lat
        self._lon = lon
        self._alt = alt

        if alt is None:
            self.adapt_temp = False

    def get_stations(
        self,
        freq: str = None,
        start: datetime = None,
        end: datetime = None,
        model: bool = True,
    ) -> pd.DataFrame:
        """
        Get list of nearby weather stations
        """

        # Get nearby weather stations
        stations = Stations()
        stations = stations.nearby(self._lat, self._lon, self.radius)

        # Guess altitude if not set
        if self._alt is None:
            self._alt = stations.fetch().head(self.max_count)["elevation"].mean()

        # Captue unfiltered weather stations
        unfiltered = stations.fetch()
        if self.alt_range:
            unfiltered = unfiltered[
                abs(self._alt - unfiltered["elevation"]) <= self.alt_range
            ]

        # Apply inventory filter
        if freq and start and end:
            age = (datetime.now() - end).days
            if model == False or age > 180:
                stations = stations.inventory(freq, (start, end))

        # Apply altitude filter
        stations = stations.fetch()
        if self.alt_range:
            stations = stations[
                abs(self._alt - stations["elevation"]) <= self.alt_range
            ]

        # Fill up stations
        selected: int = len(stations.index)
        if selected < self.max_count:
            # Remove already included stations from unfiltered
            unfiltered = unfiltered.loc[~unfiltered.index.isin(stations.index)]
            # Append to existing DataFrame
            stations = pd.concat((stations, unfiltered.head(self.max_count - selected)))

        # Score values
        if self.radius:

            # Calculate score values
            stations["score"] = (
                (1 - (stations["distance"] / self.radius)) * self.weight_dist
            ) + (
                (1 - (abs(self._alt - stations["elevation"]) / self.alt_range))
                * self.weight_alt
            )

            # Sort by score (descending)
            stations = stations.sort_values("score", ascending=False)

        # Capture result
        self._stations = stations.index[: self.max_count]

        return stations.head(self.max_count)

    @property
    def alt(self) -> int:
        """
        Returns the point's altitude
        """

        # Return altitude
        return self._alt

    @property
    def stations(self) -> pd.Index:
        """
        Returns the point's weather stations
        """

        # Return weather stations
        return self._stations



"""
Monthly Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from datetime import datetime
from typing import Union
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.interface.timeseries import TimeSeries
from meteostat.interface.point import Point


class Monthly(TimeSeries):

    """
    Retrieve monthly weather data for one or multiple weather stations or
    a single geographical point
    """

    # The cache subdirectory
    cache_subdir: str = "monthly"

    # Granularity
    granularity = Granularity.MONTHLY

    # Default frequency
    _freq: str = "1MS"

    # Flag which represents model data
    _model_flag = "I"

    # Columns
    _columns: list = [
        "year",
        "month",
        "tavg",
        "tmin",
        "tmax",
        "prcp",
        "wspd",
        "pres",
        "tsun",
    ]

    # Index of first meteorological column
    _first_met_col = 2

    # Data types
    _types: dict = {
        "tavg": "float64",
        "tmin": "float64",
        "tmax": "float64",
        "prcp": "float64",
        "wspd": "float64",
        "pres": "float64",
        "tsun": "float64",
    }

    # Columns for date parsing
    _parse_dates: dict = {"time": [0, 1]}

    # Default aggregation functions
    aggregations: dict = {
        "tavg": "mean",
        "tmin": "mean",
        "tmax": "mean",
        "prcp": "sum",
        "wspd": "mean",
        "pres": "mean",
        "tsun": "sum",
    }

    def __init__(
        self,
        loc: Union[pd.DataFrame, Point, list, str],  # Station(s) or geo point
        start: datetime = None,
        end: datetime = None,
        model: bool = True,  # Include model data?
        flags: bool = False,  # Load source flags?
    ) -> None:

        # Set start date
        if start is not None:
            start = start.replace(day=1)

        # Initialize time series
        self._init_time_series(loc, start, end, model, flags)

    def expected_rows(self) -> int:
        """
        Return the number of rows expected for the defined date range
        """

        return (
            (self._end.year - self._start.year) * 12
            + self._end.month
            - self._start.month
        ) + 1



"""
Normalize Data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
from numpy import NaN
import pandas as pd
import pytz
from meteostat.core.warn import warn


def normalize(self):
    """
    Normalize the DataFrame
    """

    if self.count() == 0:
        warn("Pointless normalization of empty DataFrame")

    # Create temporal instance
    temp = copy(self)

    if temp._start and temp._end and temp.coverage() < 1:

        # Create result DataFrame
        result = pd.DataFrame(columns=temp._columns[temp._first_met_col :])

        # Handle tz-aware date ranges
        if hasattr(temp, "_timezone") and temp._timezone is not None:
            timezone = pytz.timezone(temp._timezone)
            start = temp._start.astimezone(timezone)
            end = temp._end.astimezone(timezone)
        else:
            start = temp._start
            end = temp._end

        # Go through list of weather stations
        for station in temp._stations:
            # Create data frame
            df = pd.DataFrame(columns=temp._columns[temp._first_met_col :])
            # Add time series
            df["time"] = pd.date_range(
                start,
                end,
                freq=self._freq,
                tz=temp._timezone if hasattr(temp, "_timezone") else None,
            )
            # Add station ID
            df["station"] = station
            # Add columns
            for column in temp._columns[temp._first_met_col :]:
                # Add column to DataFrame
                df[column] = NaN

            result = pd.concat([result, df], axis=0)

        # Set index
        result = result.set_index(["station", "time"])

        # Merge data
        temp._data = (
            pd.concat([temp._data, result], axis=0)
            .groupby(["station", "time"], as_index=True)
            .first()
        )

        # None -> NaN
        temp._data = temp._data.fillna(NaN)

    # Return class instance
    return temp



"""
Utilities - Helpers

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import numpy as np


def get_distance(lat1, lon1, lat2, lon2) -> float:
    """
    Calculate distance between weather station and geo point
    """
    # Earth radius in meters
    radius = 6371000

    # Degress to radian
    lat1, lon1, lat2, lon2 = map(np.deg2rad, [lat1, lon1, lat2, lon2])

    # Deltas
    dlat = lat2 - lat1
    dlon = lon2 - lon1

    # Calculate distance
    arch = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
    arch_sin = 2 * np.arcsin(np.sqrt(arch))

    return radius * arch_sin



"""
Utilities - Validations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import pandas as pd


def validate_series(df: pd.DataFrame, station: str) -> pd.DataFrame:
    """
    Make sure a series is formatted correctly
    """

    # Add missing column(s)
    if "time" not in df.columns:
        df["time"] = None

    # Add weather station ID
    df["station"] = station

    # Set index
    df = df.set_index(["station", "time"])

    # Return DataFrame
    return df



"""
Core Class - Warnings

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import warnings


def _format(message, category, _filename, _lineno, _line=None) -> str:
    """
    Print warning on a single line
    """

    return "%s: %s\n" % (category.__name__, message)


# Set warning format
warnings.formatwarning = _format


def warn(message: str) -> None:
    """
    Create a warning
    """

    try:
        warnings.warn(message, Warning)
    except TypeError:
        pass



"""
█▀▄▀█ █▀▀ ▀█▀ █▀▀ █▀█ █▀ ▀█▀ ▄▀█ ▀█▀
█░▀░█ ██▄ ░█░ ██▄ █▄█ ▄█ ░█░ █▀█ ░█░

A Python library for accessing open weather and climate data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

__appname__ = "meteostat"
__version__ = "1.6.4"

from .interface.base import Base
from .interface.timeseries import TimeSeries
from .interface.stations import Stations
from .interface.point import Point
from .interface.hourly import Hourly
from .interface.daily import Daily
from .interface.monthly import Monthly
from .interface.normals import Normals



"""
Normals Interface Class

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
from typing import Union
from datetime import datetime
import numpy as np
import pandas as pd
from meteostat.enumerations.granularity import Granularity
from meteostat.core.warn import warn
from meteostat.interface.meteodata import MeteoData
from meteostat.interface.point import Point


class Normals(MeteoData):

    """
    Retrieve climate normals for one or multiple weather stations or
    a single geographical point
    """

    # The cache subdirectory
    cache_subdir: str = "normals"

    # Granularity
    granularity = Granularity.NORMALS

    # The list of weather Stations
    _stations: pd.Index = None

    # The first year of the period
    _start: int = None

    # The last year of the period
    _end: int = None

    # The data frame
    _data: pd.DataFrame = pd.DataFrame()

    # Columns
    _columns: list = [
        "start",
        "end",
        "month",
        "tmin",
        "tmax",
        "prcp",
        "wspd",
        "pres",
        "tsun",
    ]

    # Index of first meteorological column
    _first_met_col = 3

    # Data types
    _types: dict = {
        "tmin": "float64",
        "tmax": "float64",
        "prcp": "float64",
        "wspd": "float64",
        "pres": "float64",
        "tsun": "float64",
    }

    # Which columns should be parsed as dates?
    _parse_dates = None

    def __init__(
        self,
        loc: Union[pd.DataFrame, Point, list, str],
        start: int = None,
        end: int = None,
    ) -> None:

        # Set list of weather stations
        if isinstance(loc, pd.DataFrame):
            self._stations = loc.index

        elif isinstance(loc, Point):
            if start and end:
                stations = loc.get_stations(
                    "monthly", datetime(start, 1, 1), datetime(end, 12, 31)
                )
            else:
                stations = loc.get_stations()

            self._stations = stations.index

        else:
            if not isinstance(loc, list):
                loc = [loc]

            self._stations = pd.Index(loc)

        # Check period
        if (start and end) and (
            end - start != 29 or end % 10 != 0 or end >= datetime.now().year
        ):
            raise ValueError("Invalid reference period")

        # Set period
        self._start = start
        self._end = end

        # Get data for all weather stations
        self._data = self._get_data()

        # Interpolate data
        if isinstance(loc, Point):
            self._resolve_point(loc.method, stations, loc.alt, loc.adapt_temp)

        # Clear cache
        if self.max_age > 0 and self.autoclean:
            self.clear_cache()

    def normalize(self):
        """
        Normalize the DataFrame
        """

        # Create temporal instance
        temp = copy(self)

        if self.count() == 0:
            warn("Pointless normalization of empty DataFrame")

        # Go through list of weather stations
        for station in temp._stations:
            # The list of periods
            periods: pd.Index = pd.Index([])
            # Get periods
            if self.count() > 0:
                periods = temp._data[
                    temp._data.index.get_level_values("station") == station
                ].index.unique("end")
            elif periods.size == 0 and self._end:
                periods = pd.Index([self._end])
            # Go through all periods
            for period in periods:
                # Create DataFrame
                df = pd.DataFrame(columns=temp._columns[temp._first_met_col :])
                # Populate index columns
                df["month"] = range(1, 13)
                df["station"] = station
                df["start"] = period - 29
                df["end"] = period
                # Set index
                df.set_index(["station", "start", "end", "month"], inplace=True)
                # Merge data
                temp._data = (
                    pd.concat([temp._data, df], axis=0)
                    .groupby(["station", "start", "end", "month"], as_index=True)
                    .first()
                    if temp._data.index.size > 0
                    else df
                )

        # None -> NaN
        temp._data = temp._data.fillna(np.NaN)

        # Return class instance
        return temp

    def fetch(self) -> pd.DataFrame:
        """
        Fetch DataFrame
        """

        # Copy DataFrame
        temp = copy(self._data)

        # Add avg. temperature column
        temp.insert(
            0, "tavg", temp[["tmin", "tmax"]].dropna(how="any").mean(axis=1).round(1)
        )

        # Remove station index if it's a single station
        if len(self._stations) == 1 and "station" in temp.index.names:
            temp = temp.reset_index(level="station", drop=True)

        # Remove start & end year if period is set
        if self._start and self._end and self.count() > 0:
            temp = temp.reset_index(level="start", drop=True)
            temp = temp.reset_index(level="end", drop=True)

        # Return data frame
        return temp

    # Import methods
    from meteostat.series.convert import convert
    from meteostat.series.count import count
    from meteostat.core.cache import clear_cache



"""
Setup file

The code is licensed under the MIT license.
"""

from os import path
from setuptools import setup, find_packages

# Content of the README file
here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md")) as f:
    long_description = f.read()

# Setup
setup(
    name="meteostat",
    version="1.6.4",
    author="Meteostat",
    author_email="info@meteostat.net",
    description="Access and analyze historical weather and climate data with Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/meteostat/meteostat-python",
    keywords=["weather", "climate", "data", "timeseries", "meteorology"],
    python_requires=">=3.6.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pandas>=1.1", "pytz", "numpy"],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Database",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



"""
Fetch Data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

from copy import copy
import pandas as pd


def fetch(self) -> pd.DataFrame:
    """
    Fetch DataFrame
    """

    # Copy DataFrame
    temp = copy(self._data)

    # Remove station index if it's a single station
    if len(self._stations) == 1 and "station" in temp.index.names:
        temp = temp.reset_index(level="station", drop=True)

    # Return data frame
    return temp
